from mesido.pycml.component_library.milp._internal.heat_component import BaseAsset


class ElectricityComponent(BaseAsset):
    """
    Base electricity component nothing to add here yet.
    """

    pass
